@@ -21,6 +21,10 @@
     #endif
 #endif
 
+#ifdef HAVE_REDIS_ZSTD
+#include <zstd.h>
+#endif
+
 #include <zend_exceptions.h>
 #include "php_redis.h"
 #include "library.h"
@@ -56,18 +60,27 @@ extern int le_redis_pconnect;
 static ConnectionPool *
 redis_sock_get_connection_pool(RedisSock *redis_sock)
 {
+    ConnectionPool *p;
     zend_string *persistent_id = strpprintf(0, "phpredis_%s:%d", ZSTR_VAL(redis_sock->host), redis_sock->port);
     zend_resource *le = zend_hash_find_ptr(&EG(persistent_list), persistent_id);
-    if (!le) {
-        ConnectionPool *p = pecalloc(1, sizeof(*p) + sizeof(*le), 1);
+
+    if (le) {
+        p = le->ptr;
+    } else {
+        p = pecalloc(1, sizeof(*p), 1);
         zend_llist_init(&p->list, sizeof(php_stream *), NULL, 1);
-        le = (zend_resource *)((char *)p + sizeof(*p));
-        le->type = le_redis_pconnect;
-        le->ptr = p;
+#if (PHP_VERSION_ID < 70300)
+        zend_resource res;
+        res.type = le_redis_pconnect;
+        res.ptr = p;
+        le = &res;
         zend_hash_str_update_mem(&EG(persistent_list), ZSTR_VAL(persistent_id), ZSTR_LEN(persistent_id), le, sizeof(*le));
+#else
+        le = zend_register_persistent_resource(ZSTR_VAL(persistent_id), ZSTR_LEN(persistent_id), p, le_redis_pconnect);
+#endif
     }
     zend_string_release(persistent_id);
-    return le->ptr;
+    return p;
 }
 
 /* Helper to reselect the proper DB number when we reconnect */
@@ -105,8 +118,7 @@ redis_sock_auth(RedisSock *redis_sock)
     char *cmd, *response;
     int cmd_len, response_len;
 
-    cmd_len = redis_spprintf(redis_sock, NULL, &cmd, "AUTH", "s",
-                             ZSTR_VAL(redis_sock->auth), ZSTR_LEN(redis_sock->auth));
+    cmd_len = redis_spprintf(redis_sock, NULL, &cmd, "AUTH", "S", redis_sock->auth);
 
     if (redis_sock_write(redis_sock, cmd, cmd_len) < 0) {
         efree(cmd);
@@ -155,6 +167,8 @@ redis_error_throw(RedisSock *redis_sock)
      * Disque) */
     if (!REDIS_SOCK_ERRCMP_STATIC(redis_sock, "ERR") &&
         !REDIS_SOCK_ERRCMP_STATIC(redis_sock, "NOSCRIPT") &&
+        !REDIS_SOCK_ERRCMP_STATIC(redis_sock, "NOQUORUM") &&
+        !REDIS_SOCK_ERRCMP_STATIC(redis_sock, "NOGOODSLAVE") &&
         !REDIS_SOCK_ERRCMP_STATIC(redis_sock, "WRONGTYPE") &&
         !REDIS_SOCK_ERRCMP_STATIC(redis_sock, "BUSYGROUP") &&
         !REDIS_SOCK_ERRCMP_STATIC(redis_sock, "NOGROUP"))
@@ -562,7 +576,7 @@ union resparg {
 };
 
 /* A printf like method to construct a Redis RESP command.  It has been extended
- * to take a few different format specifiers that are convienient to phpredis.
+ * to take a few different format specifiers that are convenient to phpredis.
  *
  * s - C string followed by length as a
  * S - Pointer to a zend_string
@@ -1152,8 +1166,7 @@ static void array_zip_values_and_scores(RedisSock *redis_sock, zval *z_tab,
 
     /* replace */
     zval_dtor(z_tab);
-    ZVAL_ZVAL(z_tab, &z_ret, 1, 0);
-    zval_dtor(&z_ret);
+    ZVAL_ZVAL(z_tab, &z_ret, 0, 0);
 }
 
 static int
@@ -1198,6 +1211,9 @@ redis_mbulk_reply_zipped(INTERNAL_FUNCTION_PARAMETERS, RedisSock *redis_sock,
         } else {
             add_next_index_bool(z_tab, 0);
         }
+        if (*inbuf == TYPE_ERR) {
+            redis_sock_set_err(redis_sock, inbuf + 1, len - 1);
+        }
         return -1;
     }
     numElems = atoi(inbuf+1);
@@ -1259,7 +1275,7 @@ redis_read_stream_messages(RedisSock *redis_sock, int count, zval *z_ret
 
     /* Iterate over each message */
     for (i = 0; i < count; i++) {
-        /* Consume inner multi-bulk header, message ID itself and finaly
+        /* Consume inner multi-bulk header, message ID itself and finally
          * the multi-bulk header for field and values */
         if ((read_mbulk_header(redis_sock, &mhdr) < 0 || mhdr != 2) ||
             ((id = redis_sock_read(redis_sock, &idlen)) == NULL) ||
@@ -1459,7 +1475,7 @@ PHP_REDIS_API int
 redis_read_xinfo_response(RedisSock *redis_sock, zval *z_ret, int elements)
 {
     zval zv;
-    int i, len;
+    int i, len = 0;
     char *key = NULL, *data;
     REDIS_REPLY_TYPE type;
     long li;
@@ -1737,7 +1753,7 @@ PHP_REDIS_API void redis_debug_response(INTERNAL_FUNCTION_PARAMETERS, RedisSock
  * redis_sock_create
  */
 PHP_REDIS_API RedisSock*
-redis_sock_create(char *host, int host_len, unsigned short port,
+redis_sock_create(char *host, int host_len, int port,
                   double timeout, double read_timeout,
                   int persistent, char *persistent_id,
                   long retry_interval)
@@ -1764,6 +1780,7 @@ redis_sock_create(char *host, int host_len, unsigned short port,
 
     redis_sock->serializer = REDIS_SERIALIZER_NONE;
     redis_sock->compression = REDIS_COMPRESSION_NONE;
+    redis_sock->compression_level = 0; /* default */
     redis_sock->mode = ATOMIC;
     redis_sock->head = NULL;
     redis_sock->current = NULL;
@@ -1781,6 +1798,51 @@ redis_sock_create(char *host, int host_len, unsigned short port,
     return redis_sock;
 }
 
+static int
+redis_sock_check_liveness(RedisSock *redis_sock)
+{
+    char inbuf[4096], uniqid[64];
+    int uniqid_len;
+    smart_string cmd = {0};
+    struct timeval tv;
+    size_t len;
+
+    if (redis_sock->auth) {
+        redis_cmd_init_sstr(&cmd, 1, "AUTH", sizeof("AUTH") - 1);
+        redis_cmd_append_sstr(&cmd, ZSTR_VAL(redis_sock->auth), ZSTR_LEN(redis_sock->auth));
+    }
+
+    gettimeofday(&tv, NULL);
+    uniqid_len = snprintf(uniqid, sizeof(uniqid), "phpredis_pool:%08lx%05lx:%08lx", (long)tv.tv_sec, (long)tv.tv_usec, (long)php_rand());
+    redis_cmd_init_sstr(&cmd, 1, "ECHO", sizeof("ECHO") - 1);
+    redis_cmd_append_sstr(&cmd, uniqid, uniqid_len);
+    smart_string_0(&cmd);
+
+    if (redis_sock_write(redis_sock, cmd.c, cmd.len) < 0) {
+        smart_string_free(&cmd);
+        return FAILURE;
+    }
+    smart_string_free(&cmd);
+
+    if (redis_sock_gets(redis_sock, inbuf, sizeof(inbuf) - 1, &len) < 0) {
+        return FAILURE;
+    } else if (redis_sock->auth) {
+        if (strncmp(inbuf, "+OK", 3) != 0) {
+            return FAILURE;
+        } else if (redis_sock_gets(redis_sock, inbuf, sizeof(inbuf) - 1, &len) < 0) {
+            return FAILURE;
+        }
+    }
+    if (*inbuf != TYPE_BULK ||
+        atoi(inbuf + 1) != uniqid_len ||
+        redis_sock_gets(redis_sock, inbuf, sizeof(inbuf) - 1, &len) < 0 ||
+        strncmp(inbuf, uniqid, uniqid_len) != 0
+    ) {
+        return FAILURE;
+    }
+    return SUCCESS;
+}
+
 /**
  * redis_sock_connect
  */
@@ -1802,7 +1864,7 @@ PHP_REDIS_API int redis_sock_connect(RedisSock *redis_sock)
         schema = estrndup(address, pos - address);
         address = pos + sizeof("://") - 1;
     }
-    if (redis_sock->port < 1) {
+    if (address[0] == '/' && redis_sock->port < 1) {
         host_len = snprintf(host, sizeof(host), "unix://%s", address);
         usocket = 1;
     } else {
@@ -1826,8 +1888,8 @@ PHP_REDIS_API int redis_sock_connect(RedisSock *redis_sock)
             if (zend_llist_count(&p->list) > 0) {
                 redis_sock->stream = *(php_stream **)zend_llist_get_last(&p->list);
                 zend_llist_remove_tail(&p->list);
-                /* Check socket liveness using 0 second timeout */
-                if (php_stream_set_option(redis_sock->stream, PHP_STREAM_OPTION_CHECK_LIVENESS, 0, NULL) == PHP_STREAM_OPTION_RETURN_OK) {
+
+                if (redis_sock_check_liveness(redis_sock) == SUCCESS) {
                     redis_sock->status = REDIS_SOCK_STATUS_CONNECTED;
                     return SUCCESS;
                 }
@@ -2098,16 +2160,19 @@ PHP_REDIS_API int redis_mbulk_reply_assoc(INTERNAL_FUNCTION_PARAMETERS, RedisSoc
     zval *z_keys = ctx;
 
     if (redis_sock_gets(redis_sock, inbuf, sizeof(inbuf) - 1, &len) < 0) {
-        return -1;
+        goto failure;
     }
 
-    if(inbuf[0] != '*') {
+    if (*inbuf != TYPE_MULTIBULK) {
         if (IS_ATOMIC(redis_sock)) {
             RETVAL_FALSE;
         } else {
             add_next_index_bool(z_tab, 0);
         }
-        return -1;
+        if (*inbuf == TYPE_ERR) {
+            redis_sock_set_err(redis_sock, inbuf + 1, len - 1);
+        }
+        goto failure;
     }
     numElems = atoi(inbuf+1);
     zval z_multi_result;
@@ -2137,7 +2202,15 @@ PHP_REDIS_API int redis_mbulk_reply_assoc(INTERNAL_FUNCTION_PARAMETERS, RedisSoc
     } else {
         add_next_index_zval(z_tab, &z_multi_result);
     }
-    return 0;
+    return SUCCESS;
+failure:
+    if (z_keys != NULL) {
+        for (i = 0; Z_TYPE(z_keys[i]) != IS_NULL; ++i) {
+            zval_dtor(&z_keys[i]);
+        }
+        efree(z_keys);
+    }
+    return FAILURE;
 }
 
 /**
@@ -2186,26 +2259,60 @@ redis_pack(RedisSock *redis_sock, zval *z, char **val, size_t *val_len)
     char *buf;
     int valfree;
     size_t len;
-#ifdef HAVE_REDIS_LZF
-    char *data;
-    uint32_t res;
-    double size;
-#endif
 
     valfree = redis_serialize(redis_sock, z, &buf, &len);
     switch (redis_sock->compression) {
         case REDIS_COMPRESSION_LZF:
 #ifdef HAVE_REDIS_LZF
-            /* preserve compatibility with PECL lzf_compress margin (greater of 4% and LZF_MARGIN) */
-            size = len + MIN(UINT_MAX - len, MAX(LZF_MARGIN, len / 25));
-            data = emalloc(size);
-            if ((res = lzf_compress(buf, len, data, size)) > 0) {
-                if (valfree) efree(buf);
-                *val = data;
-                *val_len = res;
-                return 1;
+            {
+                char *data;
+                uint32_t res;
+                double size;
+
+                /* preserve compatibility with PECL lzf_compress margin (greater of 4% and LZF_MARGIN) */
+                size = len + MIN(UINT_MAX - len, MAX(LZF_MARGIN, len / 25));
+                data = emalloc(size);
+                if ((res = lzf_compress(buf, len, data, size)) > 0) {
+                    if (valfree) efree(buf);
+                    *val = data;
+                    *val_len = res;
+                    return 1;
+                }
+                efree(data);
+            }
+#endif
+            break;
+        case REDIS_COMPRESSION_ZSTD:
+#ifdef HAVE_REDIS_ZSTD
+            {
+                char *data;
+                size_t size;
+                int level;
+
+                if (redis_sock->compression_level < 1) {
+#ifdef ZSTD_CLEVEL_DEFAULT
+                    level = ZSTD_CLEVEL_DEFAULT;
+#else
+                    level = 3;
+#endif
+                } else if (redis_sock->compression_level > ZSTD_maxCLevel()) {
+                    level = ZSTD_maxCLevel();
+                } else {
+                    level = redis_sock->compression_level;
+                }
+
+                size = ZSTD_compressBound(len);
+                data = emalloc(size);
+                size = ZSTD_compress(data, size, buf, len, level);
+                if (!ZSTD_isError(size)) {
+                    if (valfree) efree(buf);
+                    data = erealloc(data, size);
+                    *val = data;
+                    *val_len = size;
+                    return 1;
+                }
+                efree(data);
             }
-            efree(data);
 #endif
             break;
     }
@@ -2217,29 +2324,51 @@ redis_pack(RedisSock *redis_sock, zval *z, char **val, size_t *val_len)
 PHP_REDIS_API int
 redis_unpack(RedisSock *redis_sock, const char *val, int val_len, zval *z_ret)
 {
-#ifdef HAVE_REDIS_LZF
-    char *data;
-    int i;
-    uint32_t res;
-#endif
-
     switch (redis_sock->compression) {
         case REDIS_COMPRESSION_LZF:
 #ifdef HAVE_REDIS_LZF
-            errno = E2BIG;
-            /* start from two-times bigger buffer and
-             * increase it exponentially  if needed */
-            for (i = 2; errno == E2BIG; i *= 2) {
-                data = emalloc(i * val_len);
-                if ((res = lzf_decompress(val, val_len, data, i * val_len)) == 0) {
-                    /* errno != E2BIG will brake for loop */
+            {
+                char *data;
+                int i;
+                uint32_t res;
+
+                errno = E2BIG;
+                /* start from two-times bigger buffer and
+                 * increase it exponentially  if needed */
+                for (i = 2; errno == E2BIG; i *= 2) {
+                    data = emalloc(i * val_len);
+                    if ((res = lzf_decompress(val, val_len, data, i * val_len)) == 0) {
+                        /* errno != E2BIG will brake for loop */
+                        efree(data);
+                        continue;
+                    } else if (redis_unserialize(redis_sock, data, res, z_ret) == 0) {
+                        ZVAL_STRINGL(z_ret, data, res);
+                    }
                     efree(data);
-                    continue;
-                } else if (redis_unserialize(redis_sock, data, res, z_ret) == 0) {
-                    ZVAL_STRINGL(z_ret, data, res);
+                    return 1;
+                }
+            }
+#endif
+            break;
+        case REDIS_COMPRESSION_ZSTD:
+#ifdef HAVE_REDIS_ZSTD
+            {
+                char *data;
+                size_t len;
+
+                len = ZSTD_getFrameContentSize(val, val_len);
+                if (len >= 0) {
+                    data = emalloc(len);
+                    len = ZSTD_decompress(data, len, val, val_len);
+                    if (ZSTD_isError(len)) {
+                        efree(data);
+                        break;
+                    } else if (redis_unserialize(redis_sock, data, len, z_ret) == 0) {
+                        ZVAL_STRINGL(z_ret, data, len);
+                    }
+                    efree(data);
+                    return 1;
                 }
-                efree(data);
-                return 1;
             }
 #endif
             break;
